/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>

#include "gimageview.h"

#include "gimv_image.h"
#include "gfileutil.h"
#include "gtkutils.h"
#include "prefs.h"
#include "fileload.h"
#include "fileutil.h"
#include "image_view.h"
#include "image_window.h"
#include "thumbnail_view.h"
#include "thumbnail_window.h"


/* for filebrowser call back functions */
typedef struct _FileSel
{
   GtkWidget   *filebrowser;
   ThumbWindow *tw;
} FileSel;


static void cb_file_load_cancel                (GtkWidget *widget, OpenFiles *files);
static void cb_file_load_stop                  (GtkWidget *widget, OpenFiles *files);
static void file_open_progress_window          (OpenFiles *files);
static void filebrowser_open_files             (FileSel *filesel, ImgWinType type);
static void cb_filebrowser_open_selected_files (GtkWidget *widget, FileSel *filesel);
static void cb_filebrowser_add_thumbnail       (GtkWidget *widget, FileSel *filesel);
static void cb_filebrowser_ok_sel              (GtkWidget *widget, FileSel *filesel);
static void cb_filebrowser_close               (GtkWidget *widget, gpointer parent);


GList *files_loader_list = NULL;
gboolean loading_stop = FALSE;

OpenFiles *
files_loader_new ()
{
   OpenFiles *files;
   gboolean reset_sensitive;

   files = g_new0 (OpenFiles, 1);
   files->filelist    = NULL;
   files->dirlist     = NULL;
   files->dirname     = NULL;
   files->window      = NULL;
   files->progressbar = NULL;
   files->thumb_load_type = LOAD_CACHE;
   files->status      = GETTING_FILELIST;
   files->now_file    = NULL;
   files->pos         = 0;
   files->num         = 0;

   reset_sensitive = !files_loader_list;
   files_loader_list = g_list_append (files_loader_list, files);

   if (reset_sensitive) {
      GList *node = ThumbWinList;

      loading_stop = FALSE;

      while (node) {
	 ThumbWindow *tw = node->data;
	 thumbwin_set_sensitive (tw, THUMB_WIN_STATUS_LOADING);
	 node = g_list_next (node);
      }
   }

   return files;
}


/*
 *  free_open_files_data:
 *     @ Free all data of OpenFiles structure
 *
 *  files : Pointer to OpenFiles structure.
 */
void
files_loader_delete (OpenFiles *files)
{
   g_list_foreach (g_list_first (files->filelist), (GFunc) g_free, NULL);
   g_list_foreach (g_list_first (files->dirlist),  (GFunc) g_free, NULL);

   g_list_free (files->filelist);
   g_list_free (files->dirlist);

   files_loader_list = g_list_remove (files_loader_list, files);

   if (!files_loader_list) {
      GList *node = ThumbWinList;
      while (node) {
	 ThumbWindow *tw = node->data;
	 thumbwin_set_sensitive (tw, THUMB_WIN_STATUS_NORMAL);
	 node = g_list_next (node);
      }
   }

   loading_stop = FALSE;

   g_free (files);
}


gboolean
files_loader_query_loading ()
{
   if (files_loader_list)
      return TRUE;
   else
      return FALSE;
}


void
files_loader_stop ()
{
   GList *node = files_loader_list;

   if (!files_loader_list) return;

   while (node) {
      OpenFiles *files = node->data;
      files->status = STOP;
      node = g_list_next (node);
   }

   get_dir_stop ();

   loading_stop = TRUE;
}


static void
cb_file_load_cancel (GtkWidget *widget, OpenFiles *files)
{
   files->status = CANCEL;
}


static void
cb_file_load_stop   (GtkWidget *widget, OpenFiles *files)
{
   files->status = STOP;
   if (files_loader_query_loading ())
      files_loader_stop ();
}


/*
 *  file_open_progress_window:
 *     @ Create popup window contains progress bar and stop button
 *       for image file open in image view window.
 *
 *  files : Pointer to OpenFiles structure (include various data for file open).
 */
static void
file_open_progress_window (OpenFiles *files)
{
   GtkWidget *window;
   GtkWidget *vbox, *hbox;
   GtkWidget *label;
   GtkWidget *progressbar;
   GtkWidget *button;

   /* create dialog window */
   window = gtk_window_new(GTK_WINDOW_DIALOG);
   gtk_container_border_width (GTK_CONTAINER (window), 3);
   gtk_window_set_title (GTK_WINDOW (window), _("*Loading Image Files* - GImageView - "));
   gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
   vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_add (GTK_CONTAINER(window), vbox);

   /* label */
   label = gtk_label_new (_("Now Opening Image Files..."));
   gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);

   /* progress bar */
   progressbar = gtk_progress_bar_new();
   gtk_progress_set_show_text(GTK_PROGRESS(progressbar), TRUE);
   gtk_box_pack_start (GTK_BOX (vbox), progressbar, FALSE, FALSE, 0);

   /* hbox */
   hbox = gtk_hbox_new (TRUE, 0);
   gtk_container_add (GTK_CONTAINER(vbox), hbox);

   /* cancel button */
   button = gtk_button_new_with_label (_("Skip"));
   gtk_container_border_width (GTK_CONTAINER (button), 5);
   gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT(button), "clicked",
		       GTK_SIGNAL_FUNC(cb_file_load_cancel), files);
 
   /* stop button */
   button = gtk_button_new_with_label (_("Stop"));
   gtk_container_border_width (GTK_CONTAINER (button), 5);
   gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT(button), "clicked",
		       GTK_SIGNAL_FUNC(cb_file_load_stop), files);


   files->window = window;
   files->progressbar = progressbar;

   gtk_widget_show_all (window);
}


/*
 *  open_image_files_in_image_view:
 *
 *  files  : Pointer to OpenFiles structure.
 *  Return : Return TRUE if success or canceled by user.
 */
gint
open_image_files_in_image_view (OpenFiles *files)
{
   gint listnum, pos;
   gint retval = FALSE;
   gfloat progress;
   gchar buf[32];
   GList *node;

   while( gtk_events_pending() ) gtk_main_iteration();

   listnum = g_list_length (g_list_first(files->filelist));

   if (listnum < 1) {
      if (conf.disp_filename_stdout)
	 g_print (_("No image files specified!!\n"));
      return FALSE;
   }

   retval = TRUE;

   if (listnum > 2)
      file_open_progress_window (files);
   else {
      files->window = NULL;
      files->progressbar = NULL;
   }

   node = files->filelist;
   while (node) {
      if (files->status == CANCEL || files->status == STOP) {
	 retval = files->status;
	 break;
      }

      /* Load Image */
      if (conf.disp_filename_stdout)
	 g_print (_("filename = %s\n"), (gchar *) node->data);

      /* open iamge window */
      imagewin_open_window(node->data);

      if (files->progressbar) {
	 gdk_window_raise (files->window->window);
	 pos = g_list_position (files->filelist, node) + 1;
	 progress = (gfloat) pos / (gfloat) listnum;
	 gtk_progress_bar_update (GTK_PROGRESS_BAR(files->progressbar), progress);
	 g_snprintf (buf, 32, "%d/%d files", pos, listnum);
	 gtk_progress_set_format_string(GTK_PROGRESS(files->progressbar), buf);
      }

      node = g_list_next (node);
   }

   if (files->window)
      gtk_widget_destroy (files->window);

   return retval;
}


static ThumbWindow *
select_thumbnail_window (ThumbWindow *tw)
{
   ThumbWindow *retval;
   GList *node;

   if (tw && g_list_find (ThumbWinList, tw)) {
      retval = (ThumbWindow *) tw;
   }
   else if (g_list_length (ThumbWinList) < 1) {
      retval = thumbwin_open_window ();
   } else {
      node = g_list_last (ThumbWinList);
      retval = node->data;
   }

   return retval;
}


/*
 *  open_image_files_in_thumbnail_view:
 *     @ Open images in thumbnail. Thumbnail view window and page will be
 *       selected automatically.
 *
 *  files  : Pointer to OpenFiles structure.
 *  tw     : Thumbnail window for open thumbnail
 *           (Pointer to ThumbWindow structure).
 *           if this value is NULL, choose other window or create it.
 *  Return : Return TRUE if success or canceled by user.
 */
gint
open_image_files_in_thumbnail_view (OpenFiles *files, ThumbWindow *tw)
{
   ThumbWindow *tw_tmp;
   ThumbView  *tv;
   GtkWidget   *newtab;

   if (g_list_length(files->filelist) < 1) {
      if (conf.disp_filename_stdout)
	 g_print (_("No image files specified!!\n"));
      return FALSE;
   }

   /* select thumbnail window */
   tw_tmp = select_thumbnail_window (tw);
   if (conf.thumbwin_raise_window)
      gdk_window_raise (tw_tmp->window->window);

   /* select notebook page and open files */
   tv = thumbwin_find_thumbtable (tw_tmp, THUMB_CURRENT_PAGE);

   if (tv && tv->mode == THUMB_MODE_COLLECTION) {
      thumbview_append_thumbnail (tv, files, FALSE);
   } else {
      if (!(tw_tmp->pagenum) || tv) {
	 newtab = thumbwin_create_new_tab (tw_tmp);
	 if (!newtab)
	    return FALSE;
      } else {
	 gint pagenum;
	 pagenum = gtk_notebook_get_current_page (GTK_NOTEBOOK(tw_tmp->notebook));
	 newtab = gtk_notebook_get_nth_page (GTK_NOTEBOOK(tw_tmp->notebook), pagenum);
      }

      /* open thumbnails */
      tv = thumbview_create (files, tw_tmp, newtab, THUMB_MODE_COLLECTION);

      if (files->status == STOP || files->status < 0) {
	 return files->status;
      } else if (!tv)
	 return FALSE;
   }

   thumbwin_set_statusbar_page_info (tw_tmp, THUMB_CURRENT_PAGE);

   return TRUE;
}


/*
 *  open_image_files:
 *     @ Open image files. Window type will be automatically selected.
 *
 *  files  : Pointer to OpenFiles structure.
 *  Return : TRUE if success or canceled by user.
 */
gint
open_image_files (OpenFiles *files)
{
   if (g_list_length(files->filelist) < 1) {
      if (conf.disp_filename_stdout)
	 g_print (_("No image files specified!!\n"));
      return FALSE;
   }

   if (conf.default_file_open_window == THUMBNAIL_WINDOW)
      return open_image_files_in_thumbnail_view (files, NULL);
   else
      return open_image_files_in_image_view (files);
}


/*
 *  open_dir_images:
 *     @ Open all image files in specified directory.
 *
 *  dirname : Directory to scan.
 *  tw      : Thumbnail window for open.
 *            If this value is NULL, select automaticaly
 *            (include use image window or not).
 *  type    : How load thumbail.
 *  gboolean: scan subdir or not.
 *  Return  : Return TRUE if success or canceled by user. 
 */
/* FIXME!! */
gint
open_dir_images (const gchar *dirname, ThumbWindow *tw,
		 ThumbLoadType type, gboolean scan_subdir)
{
   OpenFiles *files;
   ThumbWindow *tw_tmp = NULL;
   ThumbView *tv;
   GtkWidget *newtab;
   gint retval = FALSE, flags;

   tv = thumbview_find_opened_dir (dirname);
   if (tv) {
      gint pagenum;
      tw_tmp = tv->thumb_window;
      gdk_window_raise (tw_tmp->window->window);
      pagenum = gtk_notebook_page_num (GTK_NOTEBOOK (tw_tmp->notebook), tv->container);
      gtk_notebook_set_page (GTK_NOTEBOOK (tw_tmp->notebook), pagenum);
      if (!scan_subdir)
	 return TRUE;
   }

   /* allocate OpenFiles structure for this direcory */
   files = files_loader_new ();
   files->thumb_load_type = type;
   files->dirname         = dirname;

   flags = GETDIR_ENABLE_CANCEL;
   if (conf.recursive_follow_link)
      flags = flags | GETDIR_FOLLOW_SYMLINK;
   if (conf.read_dotfile)
      flags = flags | GETDIR_READ_DOT;
   if (conf.detect_filetype_by_ext)
      flags = flags | GETDIR_DETECT_EXT;
   if (conf.disp_filename_stdout)
      flags = flags | GETDIR_DISP_STDOUT;
   get_dir (dirname, flags, &files->filelist, &files->dirlist);
   if (loading_stop) goto FUNC_END;

   if (tv && scan_subdir)
      goto OPEN_SUB_DIR;

   if (!conf.thumbwin_force_open_tab && g_list_length(files->filelist) < 1) {
      if (conf.disp_filename_stdout)
	 g_print (_("No image files in this directory: %s\n"), dirname);
      goto OPEN_SUB_DIR;
   }

   files->filelist = g_list_first (files->filelist);
   if (!tw && conf.default_dir_open_window == IMAGE_VIEW_WINDOW) {
       //(!tw && conf.default_dir_open_window == THUMBNAIL_WINDOW)) {
      retval = open_image_files_in_image_view (files);
   } else {
      tw_tmp = select_thumbnail_window (tw);
      if (conf.thumbwin_raise_window)
	 gdk_window_raise (tw_tmp->window->window);

      tv = thumbwin_find_thumbtable (tw_tmp, THUMB_CURRENT_PAGE);
      if (!(tw_tmp->pagenum) || tv) {
	 newtab = thumbwin_create_new_tab (tw_tmp);
	 if (!newtab)
	    goto FUNC_END;
      } else {
	 gint pagenum;
	 pagenum = gtk_notebook_get_current_page (GTK_NOTEBOOK(tw_tmp->notebook));
	 newtab = gtk_notebook_get_nth_page (GTK_NOTEBOOK(tw_tmp->notebook), pagenum);
      }

      /* open thumbnails */
      tv = thumbview_create (files, tw_tmp, newtab, THUMB_MODE_DIR);

      /* error handling */
      if (files->status == STOP || files->status < 0) {
	 retval = files->status;
      } else if (!tv)
	 goto FUNC_END;
      else
	 retval = TRUE;
   }

   /* open directories recursively */
 OPEN_SUB_DIR:
   if (scan_subdir && g_list_length(files->dirlist) > 0
       && files->status != STOP && files->status >= 0) {
      retval = open_dirs (files, tw_tmp, type, scan_subdir);
   }

 FUNC_END:
   /* free files (allocated at head of this function) */
   files_loader_delete (files);

   return retval;
}
/* END FIXME!! */


/*
 *  open_dirs:
 *     @ Open some directories.
 *
 *  files  : Pointer to OpenFiles structure.
 *  type   : How load thumbail (from cache or create from original image).
 *  Return : Return TRUE if success or canceled by user. 
 */
gint
open_dirs (OpenFiles *files, ThumbWindow *tw,
	   ThumbLoadType type, gboolean scan_subdir)
{
   gint i;
   gint listnum;
   gint retval = FALSE;
   LoadStatus status;

   if (g_list_length (files->dirlist) > 0) {
      listnum = g_list_length (g_list_first(files->dirlist));
      for (i = 0; i < listnum; i++) {
	 status = open_dir_images (files->dirlist->data, tw, type,
				   scan_subdir);

	 if (files->status == STOP || files->status < 0
	     || status == STOP || status < 0
	     || loading_stop)
	    return files->status = STOP;
	 else
	    retval = status || retval;

	 files->dirlist = g_list_next (files->dirlist);
      }
   }

   return retval;
}


/*
 *  open_images_dirs:
 *     @ Open some images and directories.
 *     @ Open window is selected automatically.
 *
 *  files  : Pointer to OpenFiles structure.
 *  type   : How load thumbail (from cache or create from original image).
 *  Return : Return TRUE if success or canceled by user. 
 */
gint
open_images_dirs (GList *list, ThumbWindow *tw,
		  ThumbLoadType type, gboolean scan_subdir)
{
   OpenFiles *files;
   GList *node;
   gchar *path;
   LoadStatus retval = FALSE;

   files = files_loader_new ();
   files->thumb_load_type = type;

   node = list;
   while (node) {
      path = node->data;
      if (isdir (path)) {
	 files->dirlist = g_list_append (files->dirlist, g_strdup (path));
      } else if (file_exists (path)
		 && (!conf.detect_filetype_by_ext
		     || gimv_image_detect_type_by_ext (path)))
      {
	 files->filelist = g_list_append (files->filelist, g_strdup (path));
      }
      node = g_list_next (node);
   }

   if (files->filelist) {
      if (tw)
	 retval = open_image_files_in_thumbnail_view (files, tw);
      else
	 retval = open_image_files (files);
   }

   if (files->dirlist)
      retval = open_dirs (files, tw, type, scan_subdir) || retval;

   files_loader_delete (files);

   return retval;
}


/*
 *  filebrowser_open_files:
 *     @ Get selected files from filebrowser clist and open it.
 *
 *  filesel : Pointer to FileSel structure that store
 *            GtkFileSelection and thumbnail window for open.
 *  type    : Window type for open.
 */
static void
filebrowser_open_files (FileSel *filesel, ImgWinType type)
{
   GList *sel_list = NULL, *node;
   gchar *path, *path_tail, *file;
   OpenFiles *files;

   GtkFileSelection *fsel = GTK_FILE_SELECTION(filesel->filebrowser);

   /* get directory path */
   path = g_strdup(gtk_file_selection_get_filename(fsel));
   if(!path) return;
   path_tail = (gchar *) strrchr(path, '/');
   if (path_tail) *(path_tail + 1) = '\0';

   /* get selected files */
   node = GTK_CLIST(fsel->file_list)->selection;
   if (node) {
      files = files_loader_new ();
      while(node) {
	 gtk_clist_get_text(GTK_CLIST(fsel->file_list),
			    GPOINTER_TO_INT(node->data), 0, &file);
	 file = g_strconcat(path, file, NULL);
	 if (!(conf.detect_filetype_by_ext) || gimv_image_detect_type_by_ext (file))
	    files->filelist = g_list_append (files->filelist, file);
	 node = g_list_next(node);
      }
      files->filelist = g_list_first (files->filelist);
      if (type == THUMBNAIL_WINDOW) {
	 files->status = THUMB_LOADING;
	 open_image_files_in_thumbnail_view (files, filesel->tw);
      } else {
	 files->status = IMAGE_LOADING;
	 open_image_files_in_image_view (files);
      }
      files_loader_delete (files);
   }

   g_free (path);
   g_list_free(sel_list);
}


static void
cb_filebrowser_open_selected_files(GtkWidget *widget, FileSel *filesel)
{
   /* GtkFileSelection *fsel = GTK_FILE_SELECTION(filesel->filebrowser); */

   filebrowser_open_files (filesel, IMAGE_VIEW_WINDOW);

#if 0
   /* FIXME!! If filebrowser is destoryed before loading complete,
      this will cause segmentation fault */
   gtk_clist_unselect_all(GTK_CLIST (fsel->file_list));
   gtk_entry_set_text(GTK_ENTRY(fsel->selection_entry), "");
#endif
}


static void
cb_filebrowser_add_thumbnail (GtkWidget *widget, FileSel *filesel)
{
   /* GtkFileSelection *fsel = GTK_FILE_SELECTION(filesel->filebrowser); */

   filebrowser_open_files (filesel, THUMBNAIL_WINDOW);

#if 0
   /* FIXME!! If filebrowser is destoryed before loading complete,
      this will cause segmentation fault */
   gtk_clist_unselect_all(GTK_CLIST (fsel->file_list));
   gtk_entry_set_text(GTK_ENTRY (fsel->selection_entry), "");
#endif
}


/*
 *  cb_filebrowser_ok_sel:
 *     @ Callback function for filebrowser "OK" button.
 *       If specfied path is file, open in image window (Not look config).
 *       If directory, open window type is automatically ditermined.
 *
 *  widget  :
 *  filesel : User data (Pointer to GtkFileSelection thumbnail window for open).
 */
static void
cb_filebrowser_ok_sel (GtkWidget *widget, FileSel *filesel)
{
   ImageWindow *iw;
   gchar *filename;
   struct stat st;

   filename = g_strdup(gtk_file_selection_get_filename
		          (GTK_FILE_SELECTION (filesel->filebrowser)));

   if (!filename) {
      g_print (_("File name not specified!!\n"));
      return;
   }

   if (!stat(filename, &st)) {
      /* if path is directory */
      if (S_ISDIR(st.st_mode)) {
	 open_dir_images(filename, filesel->tw, LOAD_CACHE,
			 conf.scan_dir_recursive);
      /* if path is file */
      } else if (!(conf.detect_filetype_by_ext)
		 || gimv_image_detect_type_by_ext (filename))
	 {
	    if (!conf.imgwin_open_new_win && shared_img_win) {
	       imagewin_change_image_file (shared_img_win, filename);
	 } else {
	    iw = imagewin_open_window (filename);
	    if (iw && !conf.imgwin_open_new_win)
	       shared_img_win = iw;
	 }
      } else {
	 g_print (_("Not an image (or unsupported) file!!\n"));
      }
   }

   g_free (filename);
}


/*
 *  cb_filebrowser_ok_sel:
 *     @ Callback function for filebrowser "Close" button.
 *
 *  widget :
 *  parent :
 */
static void
cb_filebrowser_close (GtkWidget *widget, gpointer parent)
{
   ThumbWindow *tw = parent;

   if (tw && g_list_find (ThumbWinList, tw)) {
      tw->open_dialog = NULL;
   }
}


/*
 *  create_filebrowser:
 *     @ File open dialog (extend Gtk+'s original file open dialog widget)
 *
 *  parent : Pointer to parent window.
 *  Return : Pointer to new filebrowser.
 */
GtkWidget *
create_filebrowser (gpointer parent)
{
   GtkWidget *filebrowser, *bbox, *add_selected, *add_all, *label;
   FileSel *filesel;

   filebrowser = gtk_file_selection_new(_("Load file(s)"));
   gtk_signal_connect (GTK_OBJECT (filebrowser), "destroy",
		       GTK_SIGNAL_FUNC(cb_filebrowser_close), parent);

   filesel = g_new0 (FileSel, 1);
   filesel->filebrowser = filebrowser;
   filesel->tw          = (ThumbWindow *) parent;
   gtk_object_set_data_full (GTK_OBJECT (filebrowser), "filesel",
			     filesel, (GtkDestroyNotify) g_free);
		
   gtk_clist_set_selection_mode(GTK_CLIST(GTK_FILE_SELECTION(filebrowser)->file_list),
				GTK_SELECTION_EXTENDED);
   /*
   gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filebrowser)->selection_entry),
		      "changed", GTK_SIGNAL_FUNC(filebrowser_changed), filebrowser);
   */
   gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filebrowser)->ok_button),
                      "clicked", GTK_SIGNAL_FUNC(cb_filebrowser_ok_sel), filesel);
   gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filebrowser)->cancel_button),
			     "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT(filebrowser));

   bbox = gtk_hbutton_box_new();
   gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_END);
   gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 0);
   gtk_box_pack_end(GTK_BOX(GTK_FILE_SELECTION(filebrowser)->action_area),
		    bbox, TRUE, TRUE, 0);

   add_selected  = gtk_button_new_with_label(_("Open selected files"));
   gtk_box_pack_start(GTK_BOX(bbox), add_selected, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(add_selected), "clicked",
		      GTK_SIGNAL_FUNC(cb_filebrowser_open_selected_files), filesel);

   add_all = gtk_button_new_with_label(_("Thumbnail for selected files"));
   gtk_box_pack_start(GTK_BOX(bbox), add_all, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(add_all), "clicked",
        	      GTK_SIGNAL_FUNC (cb_filebrowser_add_thumbnail), filesel);
   gtk_widget_show_all(bbox);

   /*
    * Change the Cancel buttons caption to Close.
    */
   label = gtk_label_new(_("Close"));
   gtk_misc_set_alignment(GTK_MISC(label), 0.5, 0.5);
   gtk_container_remove(GTK_CONTAINER(GTK_FILE_SELECTION(filebrowser)->cancel_button),
      gtk_container_children(GTK_CONTAINER(GTK_FILE_SELECTION(filebrowser)->cancel_button))->data);
   gtk_container_add(GTK_CONTAINER(GTK_FILE_SELECTION(filebrowser)->cancel_button), label);
   gtk_widget_show(label);

   gtk_widget_show(filebrowser);

   return filebrowser;
}



#ifdef GIMV_DEBUG
void
file_disp_loading_status (OpenFiles *files)
{
   g_print ("Loading ID: %p   ", files);
   g_print ("status ID: %d   ", files->status);

   switch (files->status) {
      case NO_OPEN_FILE:
	 g_print ("Loading status: NO_OPEN_FILE\n");
	 break;
      case END:
	 g_print ("Loading status: END\n");
	 break;
      case GETTING_FILELIST:
	 g_print ("Loading status: GETTING_FILELIST\n");
	 break;
      case IMAGE_LOADING:
	 g_print ("Loading status: IMAGE_LOADING\n");
	 break;
      case IMAGE_LOAD_DONE:
	 g_print ("Loading status: IMAGE_LOAD_DONE\n");
	 break;
      case THUMB_LOADING:
	 g_print ("Loading status: THUMB_LOADING\n");
	 break;
      case THUMB_LOAD_DONE:
	 g_print ("Loading status: THUMB_LOAD_DONE\n");
	 break;
      case CANCEL:
	 g_print ("Loading status: CANCEL\n");
	 break;
      case STOP:
	 g_print ("Loading status: STOP\n");
	 break;
      case CONTAINER_DESTROYED:
	 g_print ("Loading status: CONTAINER_DESTROYED\n");
	 break;
      case WINDOW_DESTROYED:
	 g_print ("Loading status: WINDOW_DESTROYED\n");
	 break;
      default:
	 break;
   }
}
#endif /* GIMV_DEBUG */
